﻿--region create schema
IF SCHEMA_ID('dlv') IS NULL
BEGIN
	EXEC ('CREATE SCHEMA dlv');
END
GO
--endregion

--region dlv.tsp_Execute
IF OBJECT_ID('dlv.tsp_Execute') IS NOT NULL
	DROP PROCEDURE dlv.tsp_Execute;
GO

CREATE PROCEDURE dlv.tsp_Execute (
	@Cmd NVARCHAR(MAX),
	@NeedLog BIT = 0
)
AS BEGIN
	IF @NeedLog = 1 PRINT @Cmd;
	BEGIN TRY
		EXEC (@Cmd);
	END TRY
	BEGIN CATCH
		PRINT 'Error "' + ERROR_MESSAGE() + '"';
		PRINT 'in query "' + @Cmd + '"';
	END CATCH
END
GO
--endregion

--region dlv.fn_GetEntityColumns
IF OBJECT_ID('dlv.fn_GetEntityColumns') IS NOT NULL
	DROP FUNCTION dlv.fn_GetEntityColumns;
GO

CREATE FUNCTION dlv.fn_GetEntityColumns ( 
	@EntityName SYSNAME
)
RETURNS @EntityColumns TABLE(
	Name NVARCHAR(MAX),
	Type NVARCHAR(MAX)
)
BEGIN
	INSERT INTO @EntityColumns
	SELECT
		c.name,
		CASE
			WHEN t.name IN ('varchar', 'nvarchar') AND c.max_length = -1 THEN CONCAT(t.name, '(max)')
			WHEN t.name = 'varchar' THEN CONCAT(t.name, '(', c.max_length, ')')
			WHEN t.name = 'nvarchar' THEN CONCAT(t.name, '(', c.max_length / 2, ')')
			ELSE t.name
		END AS Type
	FROM sys.columns c
	JOIN sys.types t ON c.user_type_id = t.user_type_id
	WHERE c.object_id = OBJECT_ID(@EntityName)
		AND c.is_identity = 0;
	RETURN;
END
GO
--endregion

--region dlv.fn_GetBaseEntityColumns
IF OBJECT_ID('dlv.fn_GetBaseEntityColumns') IS NOT NULL
	DROP FUNCTION dlv.fn_GetBaseEntityColumns;
GO

CREATE FUNCTION dlv.fn_GetBaseEntityColumns ()
RETURNS @BaseEntityColumns TABLE(
	Name NVARCHAR(MAX),
	Type NVARCHAR(MAX)
)
BEGIN
	INSERT INTO @BaseEntityColumns
	SELECT 'Id',				'uniqueidentifier'	UNION
	SELECT 'ModifiedOn',		'datetime2'			UNION
	SELECT 'CreatedOn',			'datetime2'			UNION
	SELECT 'CreatedById',		'uniqueidentifier'	UNION
	SELECT 'ModifiedById',		'uniqueidentifier'	UNION
	SELECT 'ProcessListeners',	'int'
	RETURN;
	-- исследовать: для Lcz-таблиц исключать ли поля RecordId, SysCultureId ?
END
GO
--endregion

--region dlv.tsp_PrepareEntityToIgnore
IF OBJECT_ID('dlv.tsp_PrepareEntityToIgnore') IS NOT NULL
	DROP PROCEDURE dlv.tsp_PrepareEntityToIgnore;
GO
CREATE PROCEDURE dlv.tsp_PrepareEntityToIgnore
AS BEGIN
	SET NOCOUNT ON;

	IF OBJECT_ID('dlv.EntityToIgnore') IS NULL
		CREATE TABLE dlv.EntityToIgnore (
			Name SYSNAME NOT NULL
		);

	DECLARE @DefaultEntityToIgnore TABLE (
		Name NVARCHAR(250)
	);
	INSERT INTO @DefaultEntityToIgnore
		(Name)
	VALUES
		('CommunicationType'), ('ExcelImportLog'), ('HeaderProperty'), ('SysSchema'), ('SysProcessLog'),
		('KnowledgeBaseTagV2'), ('ActivityTag'), ('CaseTag'), ('AccountTag'), ('OpportunityTag'), ('LeadTag'), ('ContactTag'),
		('InvoiceFolder'), ('OpportunityFolder'), ('AccountFolder'), ('ContactFolder'), ('LeadFolder'),
		('LevelPartnershipParam'), ('MktgActivity'), ('AccountBillingInfo'), ('Fund'), ('Applicant'), ('ServicePact'), ('ConfItem'), ('Contact'),
		('OrderExpense'), ('OrgStructureUnit'), ('Product'), ('Bank'), ('Partnership'), ('Employee'), ('Payments'), ('InternalRequest'),
		('InvoiceProduct'), ('OrderProduct'), ('CompetitorProduct'), ('RecruitmentInStage'), ('SupplyPaymentElement'), ('ActivityFile'), ('Activity'), ('EmailMessageData'), ('ActivityParticipant'), ('SocialSubscription'), ('Account');

	MERGE dlv.EntityToIgnore AS target
	USING (SELECT Name FROM @DefaultEntityToIgnore) AS source
	ON target.Name = source.Name
	WHEN NOT MATCHED THEN INSERT (Name) VALUES (source.Name);
END
GO
--endregion

--region dlv.tsp_PrepareEntityColumns
IF OBJECT_ID('dlv.tsp_PrepareEntityColumns') IS NOT NULL
	DROP PROCEDURE dlv.tsp_PrepareEntityColumns;
GO

CREATE PROCEDURE dlv.tsp_PrepareEntityColumns
AS BEGIN
	SET NOCOUNT ON;

	IF OBJECT_ID('dlv.EntityToSave') IS NOT NULL
		DROP TABLE dlv.EntityToSave;

	CREATE TABLE dlv.EntityToSave (
		Name SYSNAME NOT NULL,
		ColumnName SYSNAME
	);

	DECLARE @EntityName SYSNAME;

	DECLARE cur CURSOR FAST_FORWARD READ_ONLY LOCAL FOR
		WITH cte AS (
			SELECT DISTINCT
				ss.Name,
				CASE
					WHEN ss.Name = 'SupplyPaymentTemplateItem' THEN 'SysSupplyPaymTemplItemLcz'
					WHEN ss.Name = 'OpportunityParticipantRole' THEN 'SysOpportunityParticipRoleLcz'
					WHEN ss.Name = 'EmailTemplateRecipientType' THEN 'SysEmailTemplRecipientTypeLcz'
					WHEN ss.Name = 'SynchronizationProcessState' THEN 'SysSyncProcessStateLcz'
					WHEN ss.Name = 'GlobalDuplicateSearchStatus' THEN 'SysGlobalDuplSearchStatusLcz'
					WHEN ss.Name = 'SysEntitySchemaColRightLevel' THEN 'SysEntitySchColRightLevelLcz'
					WHEN ss.Name = 'EntitySchemaRecRightOperation' THEN 'SysEntitySchRecRightOprLcz'
					WHEN ss.Name = 'SysEntitySchemaParentAsscType' THEN 'SysEntitySchParentAsscTypeLcz'
					WHEN ss.Name = 'SysEntitySchemaRecOprRightLvl' THEN 'SysEntitySchRecOprRightLvlLcz'
					WHEN ss.Name = 'SysEntitySchemaRecRightSource' THEN 'SysEntitySchRecRightSourceLcz'
					WHEN ss.Name = 'MailboxForIncidentRegistration' THEN 'SysMailboxForIncidentRegisLcz'
					WHEN LEFT(ss.Name, 3) = 'Sys' THEN ss.Name + 'Lcz'
					ELSE 'Sys' + ss.Name + 'Lcz'
				END AS LczName
			FROM SysSchema ss
			JOIN Lookup l ON ss.UId = l.SysEntitySchemaUId
			JOIN INFORMATION_SCHEMA.TABLES ist ON ist.TABLE_NAME = ss.Name
			JOIN INFORMATION_SCHEMA.columns isc ON isc.TABLE_NAME = ss.Name
			LEFT JOIN dlv.EntityToIgnore ignore ON ignore.Name = ss.Name
			WHERE ist.TABLE_TYPE <> 'VIEW'
				AND isc.TABLE_SCHEMA = 'dbo'
				AND isc.COLUMN_NAME = 'Name'
				AND ignore.Name IS NULL
			UNION SELECT 'DuplicatesRule', 'SysDuplicatesRuleLcz' FROM SysSchema
			UNION SELECT 'MailServer', 'SysMailServerLcz' FROM SysSchema
			UNION SELECT 'QualifyStatusDecoupling', NULL FROM SysSchema
		)
		SELECT cte.Name AS EntityName FROM cte UNION
		SELECT cte.LczName AS EntityName FROM cte
		WHERE OBJECT_ID(cte.LczName) IS NOT NULL;
	OPEN cur;

	FETCH NEXT FROM cur INTO @EntityName;
	WHILE @@FETCH_STATUS = 0 BEGIN
		INSERT INTO dlv.EntityToSave
			(Name, ColumnName)
		SELECT @EntityName AS Name, col.Name AS ColumnName
		FROM dlv.fn_GetEntityColumns(@EntityName) col
		EXCEPT
		SELECT @EntityName AS Name, col.Name AS ColumnName
		FROM dlv.fn_GetBaseEntityColumns() col;
		FETCH NEXT FROM cur INTO @EntityName;
	END

	CLOSE cur;
	DEALLOCATE cur;
END
GO
--endregion

--region dlv.tsp_SaveEntityColumns
IF OBJECT_ID('dlv.tsp_SaveEntityColumns') IS NOT NULL
	DROP PROCEDURE dlv.tsp_SaveEntityColumns;
GO

CREATE PROCEDURE dlv.tsp_SaveEntityColumns
AS BEGIN
	SET NOCOUNT ON;

	DECLARE @CmdTpl NVARCHAR(MAX) =
'IF OBJECT_ID(''dlv.[##Name##_backup]'') IS NOT NULL
	DROP TABLE dlv.[##Name##_backup];

IF OBJECT_ID(''[##Name##]'') IS NOT NULL
	SELECT
		*
	INTO dlv.[##Name##_backup]
	FROM [##Name##];';
	DECLARE @Cmd NVARCHAR(MAX);

	DECLARE @Name SYSNAME;

	DECLARE cur CURSOR FAST_FORWARD READ_ONLY LOCAL FOR
		SELECT DISTINCT Name
		FROM dlv.EntityToSave;

	OPEN cur;
	FETCH NEXT FROM cur INTO @Name;

	WHILE @@FETCH_STATUS = 0 BEGIN
		SET @Cmd = REPLACE(@CmdTpl, '##Name##', @Name);
		EXEC dlv.tsp_Execute @Cmd;
		FETCH NEXT FROM cur INTO @Name;
	END

	CLOSE cur;
	DEALLOCATE cur;

END
GO
--endregion

--region dlv.tsp_RestoreEntityColumns
IF OBJECT_ID('dlv.tsp_RestoreEntityColumns') IS NOT NULL
	DROP PROCEDURE dlv.tsp_RestoreEntityColumns;
GO

CREATE PROCEDURE dlv.tsp_RestoreEntityColumns
AS BEGIN
	SET NOCOUNT ON;

	IF OBJECT_ID('dlv.EntityToSave') IS NULL
		RETURN;

	DECLARE @CmdTpl NVARCHAR(MAX) =
'IF (OBJECT_ID(''[##Name##]'') IS NOT NULL) AND (OBJECT_ID(''dlv.[##Name##_backup]'') IS NOT NULL)
BEGIN
	UPDATE entity
	SET ##ColumnSetCmd##
	FROM dlv.[##Name##_backup] bak
	INNER JOIN [##Name##] entity ON bak.Id = entity.Id;

	--DROP TABLE dlv.[##Name##_backup];
END'
	DECLARE @Cmd NVARCHAR(MAX);
	DECLARE @ColumnTpl NVARCHAR(MAX) = CHAR(9) + CHAR(9) + ',entity.[##ColumnName##] = bak.[##ColumnName##]';
	DECLARE @ColumnCmd NVARCHAR(MAX);
	DECLARE @ColumnSetCmd NVARCHAR(MAX);

	DECLARE @Name SYSNAME;
	DECLARE @ColumnName SYSNAME;

	DECLARE entityCursor CURSOR FAST_FORWARD READ_ONLY LOCAL FOR
		SELECT DISTINCT Name
		FROM dlv.EntityToSave;

	OPEN entityCursor;
	FETCH NEXT FROM entityCursor INTO @Name;

	WHILE @@FETCH_STATUS = 0 BEGIN
		SET @ColumnSetCmd = '';
		DECLARE columnCursor CURSOR FAST_FORWARD READ_ONLY LOCAL FOR
			SELECT
				es.ColumnName
			FROM dlv.EntityToSave es
			JOIN INFORMATION_SCHEMA.COLUMNS isc ON
				isc.TABLE_SCHEMA = 'dbo' AND
				isc.TABLE_NAME = @Name AND
				isc.COLUMN_NAME = es.ColumnName
			WHERE es.Name = @Name;
		OPEN columnCursor;
		FETCH NEXT FROM columnCursor INTO @ColumnName;
		WHILE @@FETCH_STATUS = 0 BEGIN
			SET @ColumnCmd = REPLACE(@ColumnTpl, '##ColumnName##', @ColumnName);
			SET @ColumnSetCmd = CONCAT(@ColumnSetCmd, CHAR(13), CHAR(10), @ColumnCmd);
			FETCH NEXT FROM columnCursor INTO @ColumnName;
		END
		CLOSE columnCursor;
		DEALLOCATE columnCursor;
		SET @ColumnSetCmd = STUFF(@ColumnSetCmd, 5, 1, '');

		SET @Cmd = REPLACE(@CmdTpl, '##Name##', @Name);
		SET @Cmd = REPLACE(@Cmd, '##ColumnSetCmd##', @ColumnSetCmd);
		EXEC dlv.tsp_Execute @Cmd;
		FETCH NEXT FROM entityCursor INTO @Name;
	END

	CLOSE entityCursor;
	DEALLOCATE entityCursor;
END
GO
--endregion
